#ifndef lint
static char vcid[] = "$Id$";
#endif

/*
   This file contains the code to read the initial node list.  Recall the
   specification from the paper:

    nodelist -> '-all'
    nodelist -> domain ':' nodelist
    nodelist -> range [ , nodelist ]
    nodelist -> nodenum [ , nodelist ]
    nodelist -> nodename [ , nodelist ]
    domain   -> <predefinedname>
    domain   -> '@' <valid filename>
    range    -> nodenum '-' nodenum
    nodenum  -> <integer>
    nodenum  -> 'last'
    nodename -> <any valid nodename>

 */

#include "sut.h"

/*
   Convert a list of ranges into a list of individual numbers
 */
int SUT_Get_host_list( ranges, nranges, host_nums, count )
SUT_Range *ranges;
int       nranges, **host_nums, *count;
{
int currange, c, tot, *idx;
tot = 0;
for (currange = 0; currange < nranges; currange++) {
    tot += (1 + ranges[currange].end - ranges[currange].start) / 
	      ranges[currange].stride;
    }
if (tot <= 0) return SUT_RANGES_INVALID;
*host_nums = idx = (int *)MALLOC( tot * sizeof(int) );
*count     = tot;
if (!idx) return SUT_NO_MEMORY;
for (currange = 0; currange < nranges; currange++) {
    for (c = ranges[currange].start; c <= ranges[currange].end; 
	 c += ranges[currange].stride) 
	*idx++ = c;
    }
return SUT_SUCCESS;
}

/*
 Rotate the hostnum list so that I am first, if I'm in the list 
 */
int SUT_Rotate_host_list( host_nums, count )
int *host_nums, count;
{
int i, j, myid;

myid = SUT_Mynodenum();
if (myid < 0) return SUT_SUCCESS;

if (host_nums[0] == myid) return SUT_SUCCESS;
for (i=1; i<count; i++) {
    if (host_nums[i] == myid) {
	int *itmp = (int *)MALLOC( count * sizeof(int) );
	if (!itmp) return SUT_NO_MEMORY;
	for (j=0; j<count; j++) 
	    itmp[j] = host_nums[(i+j) % count];
	for (j=0; j<count; j++) 
	    host_nums[j] = itmp[j];
	FREE( itmp );
	break;
	}
    }
return SUT_SUCCESS;
}
/* 
   Get the node list from the command line.  If not a simple range in the
   default domain, hctx will be set to convert the host_nums to specific 
   node names.

   This is obsolete; set SUT_Get_node_list 

 */
int SUT_Get_node_list( argc, argv, nhost, host_nums, hctx )
int  argc;
char **argv;
int  *nhost, **host_nums;
void **hctx;
{
int         err;
SUT_Range   *ranges;
int         nranges;


if (err = SUT_Get_ranges( argc, argv, &ranges, &nranges )) {
    fprintf( stderr, "Could not get process ranges from command line\n" );
    SUT_Errormessage( stderr, err );
    exit(1);
    }
/* Get the number and list of node NUMBERS */
if (err = SUT_Get_host_list( ranges, nranges, host_nums, nhost )) {
    SUT_Errormessage( stderr, err );
    exit(1);
    }
*hctx = 0;
return SUT_SUCCESS;
}

/* internal; used to create the final list */
typedef struct _nodelist {
    int       type;
    struct _nodelist *next, *tail;
    union {
        SUT_Range  range;
        char       *name;
	SUT_Domain domain;
        } x;
    } SUT_Nodelist;

/*
    nodelist -> '-all'
    nodelist -> domain ':' nodelist
    nodelist -> range [ , nodelist ]
    nodelist -> nodenum [ , nodelist ]
    nodelist -> nodename [ , nodelist ]
    domain   -> <predefinedname>
    domain   -> '@' <valid filename>
    range    -> nodenum '-' nodenum
    nodenum  -> <integer>
    nodenum  -> 'last'
    nodename -> <any valid nodename>
 */

/*
    Here are the routines to read the list
 */

char *SUT_Readdomain();
char *SUT_ReadRange();
char *SUT_Readname();

int SUT_ReadNodelist( str, list )
char         *str;
SUT_Nodelist *list;
{
while (str && *str) {
    if (*str == ',')
	str++;
    else if (*str == '@')
	str = SUT_Readdomain( str+1, list );
    else if (isdigit(*str))
	str = SUT_ReadRange( str, list );
    else 
	str = SUT_Readname( str, list );
    }
return SUT_SUCCESS;
}

#define SUT_MAX_NAME 64
#define SUT_MAX_NP   256
char *SUT_Readdomain( str, list )
char *str;
SUT_Nodelist *list;
{
char *fname, *p;
char buf[SUT_MAX_NAME];
FILE *domain;
char **names;
int  np, i;
SUT_Nodelist *nlist;

/* get the filename */
fname = p = str;
while (*p && *p != ':') p++;
if (*p) str = p + 1;
else    str = p;
*p = 0;
domain = fopen( fname, "r" );
if (!domain) {
    fprintf( stderr, "Could not open domain %s\n", fname );
    return 0;  /* FIX */
    }

/* Read domain into internal structure */
np = 0;
names = (char **)MALLOC( SUT_MAX_NP * sizeof(char *) );
if (!names) return 0;

while (1) {
    if (!fgets( buf, SUT_MAX_NAME, domain )) break;
    /* Strip the newline */
    if (buf[strlen(buf)-1] == '\n')
	buf[strlen(buf)-1] = 0;
    names[np] = (char *)MALLOC( strlen(buf) + 1 );
    if (!names[np]) return 0;
    strcpy( names[np], buf );
    np++;
    if (np > SUT_MAX_NP) 
	return 0;
    }

nlist = (SUT_Nodelist *)MALLOC( sizeof(SUT_Nodelist) );
if (!nlist) return 0;
nlist->type	 = 3;
nlist->next      = 0;
list->tail->next = nlist;
list->tail	 = nlist;
nlist->x.domain.n = np;
nlist->x.domain.domain_name = (char *)MALLOC( strlen(fname) + 1 );
if (!nlist->x.domain.domain_name) return 0;
strcpy( nlist->x.domain.domain_name, fname );
nlist->x.domain.names = (char **)MALLOC( np * sizeof(char *) );
if (!nlist->x.domain.names)
    return 0;
for (i=0; i<np; i++) 
    nlist->x.domain.names[i] = names[i];
FREE( names );

return str;
}

char *SUT_Readname( str, list )
char *str;
SUT_Nodelist *list;
{
SUT_Nodelist *nlist;
char *name, *p;

/* get the nodename */
if (*str == ',') str++;
name = p = str;
while (*p && *p != ',') p++;
if (*p) str = p + 1;
else    str = p;
*p = 0;

nlist = (SUT_Nodelist *)MALLOC( sizeof(SUT_Nodelist) );
if (!nlist) return 0;
nlist->next      = 0;
list->tail->next = nlist;
list->tail	 = nlist;
if (strcmp( name, "-all" ) == 0) {
    nlist->type = 1;
    nlist->x.range.start = 1;
    nlist->x.range.end   = 0;  /* need to fix from curdomain */
    }
else {
    nlist->type	 = 2;
    nlist->x.name     = (char *)MALLOC( strlen(name) + 1 );
    if (!nlist->x.name) return 0;
    strcpy( nlist->x.name, name );
    }
return str;
}

char *SUT_ReadRange( str, list )
char *str;
SUT_Nodelist *list;
{
char *SUT_Get_single_range();
SUT_Nodelist *nlist;

nlist = (SUT_Nodelist *)MALLOC( sizeof(SUT_Nodelist) );
if (!nlist) return 0;
nlist->type	 = 1;
nlist->next      = 0;
list->tail->next = nlist;
list->tail	 = nlist;
return SUT_Get_single_range( str, &nlist->x.range );
}

/*
   Get the ranges from the argument list
 */
int SUT_Get_ranges( argc, argv, ranges, nranges )
int argc;
char **argv;
SUT_Range **ranges;
int       *nranges;
{
char *range_vec;
char *nchar, save_char;
int  s, e, nrange;
static SUT_Range lranges[100];

*nranges = 0;
if (argc < 1) return 0;
range_vec = argv[1];

/* Format to begin with is 
    RANGE = -all
    RANGE = TERM
    TERM = TERM [ , TERM]
    TERM = n-m
    TERM = n
    TERM = 
 */
nrange = 0;
if (strcmp(range_vec, "-all") == 0) {
    *nranges	      = 1;
    *ranges	      = lranges;
    lranges[0].start  = 1;
    lranges[0].end    = SUT_Num_hosts();
    lranges[0].stride = 1;
    return SUT_SUCCESS;
    }

/* We need to allow a trailing comma ... */
while (*range_vec) {
    nchar     = range_vec;
    while (isdigit(*nchar)) nchar++;
    if (nchar == range_vec) 
	return SUT_RANGES_INVALID;
    save_char = *nchar;
    *nchar    = 0;
    s	      = atoi(range_vec);
    *nchar    = save_char;
    e	      = s;
    if (save_char == '-') {
	range_vec = ++nchar;
	while (isdigit(*nchar)) nchar ++;
	save_char = *nchar;
	*nchar	  = 0;
	e	  = atoi(range_vec);
	*nchar	  = save_char;
	save_char = '-';
	}
    else if (save_char == ',') nchar++;
    if (save_char == ',' || save_char == '-' || !save_char) {
	lranges[nrange].start	 = s;
	lranges[nrange].end	 = e;
	lranges[nrange++].stride = 1;
	}
    else {
	return SUT_INVALID_RANGE_SPEC;
	}
    range_vec = nchar;
    }
*ranges  = lranges;
*nranges = nrange;
return 0;
}

/*
   Get the ranges from the argument list
 */
char *SUT_Get_single_range( str, range )
char *str;
SUT_Range *range;
{
char *range_vec;
char *nchar, save_char;
int  s, e, nrange;

range_vec = str;

/* Format to begin with is 
    RANGE = -all
    RANGE = TERM
    TERM = TERM [ , TERM]
    TERM = n-m
    TERM = n
    TERM = 
 */
if (strcmp(range_vec, "-all") == 0) {
    range->start  = 1;
    range->end    = SUT_Num_hosts();
    range->stride = 1;
    return SUT_SUCCESS;
    }

/* We need to allow a trailing comma ... */
nchar     = range_vec;
while (isdigit(*nchar)) nchar++;
if (nchar == range_vec) 
    /* err = SUT_RANGES_INVALID */
    return 0;
save_char = *nchar;
*nchar    = 0;
s	      = atoi(range_vec);
*nchar    = save_char;
e	      = s;
if (save_char == '-') {
    range_vec = ++nchar;
    while (isdigit(*nchar)) nchar ++;
    save_char = *nchar;
    *nchar	  = 0;
    e	  = atoi(range_vec);
    *nchar	  = save_char;
    save_char = '-';
    }
else if (save_char == ',') nchar++;
if (save_char == ',' || save_char == '-' || !save_char) {
    range->start	 = s;
    range->end	 = e;
    range->stride = 1;
    }
else {
    /* err = SUT_INVALID_RANGE_SPEC */
    return 0;
    }
range_vec = nchar;
return range_vec;
}

int SUT_PrintNodelist( list )
SUT_Nodelist *list;
{
SUT_Domain *curdomain = 0;
int        i;
while (list) {
    switch (list->type) {
	case 0: /* head, do nothing */
	break;
	case 1: /* range */
	if (!curdomain)
	    fprintf( stderr, "%d-%d\n", 
		     list->x.range.start, list->x.range.end );
	else
	    for (i=list->x.range.start; i<=list->x.range.end; i++) {
		fprintf( stderr, "%s\n", curdomain->names[i-1] );
		}
	break;
	case 2: /* node */
	fprintf( stderr, "%s\n", list->x.name );
	break;
	case 3: /* domain */
	curdomain = &list->x.domain;
	break;
	default:
	fprintf( stderr, "Unknown node list type %d\n", list->type );
	break;
	}
    list = list->next;
    }
return SUT_SUCCESS;
}

/* Create an array of node entries given a list */
int SUT_Create_node_list( list, np, Names )
SUT_Nodelist   *list;
int            *np;
SUT_Node_entry **Names;
{
SUT_Nodelist *cur;
SUT_Node_entry *names;
SUT_Domain *curdomain = 0;
int        i;
int        count = 0;
void       *ctx = 0;

/* First pass: get the number of processes */
cur = list;
while (cur) {
    switch (cur->type) {
	case 0: /* head, do nothing */
	break;
	case 1: /* range */
	count += (list->x.range.end - list->x.range.start + 1);
	break;
	case 2: /* node */
	count ++;
	break;
	case 3: /* domain */
	break;
	default:
	fprintf( stderr, "Unknown node list type %d\n", list->type );
	break;
	}
    cur = cur->next;
    }

*Names = names = (SUT_Node_entry *) MALLOC( count * sizeof(SUT_Node_entry) );
if (!names) return;

count = 0;
while (list) {
    switch (list->type) {
	case 0: /* head, do nothing */
	break;
	case 1: /* range */
	if (!curdomain)
	    for (i=list->x.range.start; i<=list->x.range.end; i++) {
		names[count].name    = SUT_HostName( i, ctx );
		names[count].nodenum = i;
		names[count++].domain  = 0;
		}
	else
	    for (i=list->x.range.start; i<=list->x.range.end; i++) {
		names[count].name     = curdomain->names[i-1];
		names[count].nodenum  = i;
		names[count++].domain = curdomain;
		}
	break;
	case 2: /* node */
	names[count].name    = list->x.name;
	names[count].nodenum = -1;
	names[count++].domain  = 0;
	break;
	case 3: /* domain */
	curdomain = &list->x.domain;
	break;
	default:
	fprintf( stderr, "Unknown node list type %d\n", list->type );
	break;
	}
    list = list->next;
    }
*np = count;
return SUT_SUCCESS;
}

int SUT_Print_node_entries( np, Names )
int            np;
SUT_Node_entry *Names;
{
int i;
for (i=0; i<np; i++) {
    if (Names[i].domain) 
	fprintf( stderr, "%s @ %s = %d\n", Names[i].name, 
		 Names[i].domain->domain_name, Names[i].nodenum );
    else 
	fprintf( stderr, "%s = %d\n", Names[i].name, Names[i].nodenum );
    }
}

/****************************************************************************/
/*
   Sending a node list requires sending

   The names
   The corresponing node numbers
   The domains (at least the names)

   A special case that we should eventually support is a range (or series of
   ranges) in default domain.

   The format of this will probably change over time, which is why the
   exchange is done with the SUT_Send/Recv_node_list routines.

   Currently, this does not handle the domains
 */
/****************************************************************************/

/* Send a node list to another process */
int SUT_Send_node_list( fd, entries, np )
int fd;
SUT_Node_entry *entries;
int np;
{
SUT_Header msg;
int        i;
int        len;
char       *buf, *p;

len = 0;
for (i=0; i<np; i++) 
    len += strlen( entries[i].name ) + 1 + sizeof(int);

buf = (char *)MALLOC( len + sizeof(int) );
if (!buf)
    return SUT_NO_MEMORY;
p   = buf + sizeof(int);
*(int *)buf = np;
for (i=0; i<np; i++) {
    /* We use memcpy to avoid alignment problems */
    memcpy( p, entries[i].nodenum, sizeof(int) );
    p += sizeof(int);
    strcpy( p, entries[i].name );
    p += strlen( entries[i].name) + 1; 
    }

msg.op	= SUT_NODELIST;
msg.len	= len + sizeof(int);
WRITE( fd, &msg, sizeof(SUT_Header) );
WRITE( fd, buf, len + sizeof(int) );
FREE( buf );

return SUT_SUCCESS;
}

/* Receive a node list */
int SUT_Recv_node_list( fd, np, entries )
int fd;
int *np;
SUT_Node_entry **entries;
{
SUT_Header msg;
int        len, i;
char       *buf, *p;

READ( fd, &msg, sizeof(SUT_Header) );
buf = (char *)MALLOC( msg.len );
READ( fd, buf, msg.len );
*np = *(int *)buf;
*entries = (SUT_Node_entry *) MALLOC( *np * sizeof(SUT_Node_entry) );
p   = buf + sizeof(int);
for (i=0; i<*np; i++) {
    (*entries)[i].domain   = 0;
    memcpy((*entries)[i].nodenum, p, sizeof(int) );
    p += sizeof(int);
    (*entries)[i].name	   = MALLOC( strlen(p) + 1 );
    strcpy( (*entries)[i].name, p );
    p += strlen(p) + 1;
    }
FREE( buf );
return SUT_SUCCESS;
}

/* Sample program for testing */
#ifdef SUT_NODELIST_TEST
main( argc, argv )
int argc;
char **argv;
{
SUT_Nodelist   head;
int            np;
SUT_Node_entry *names;

head.type = 0;
head.next = 0;
head.tail = &head;
SUT_ReadNodelist( argv[1], &head );

SUT_PrintNodelist( &head );

SUT_Create_node_list( &head, &np, &names );

SUT_Print_node_entries( np, names );
}
#endif
